"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OnNavigationEventGenerator = exports.shouldIgnoreElement = exports.FormSubmitGenerator = exports.ClickSignalsGenerator = exports.parseElement = void 0;
const detect_url_change_1 = require("../../../lib/detect-url-change");
const factories_1 = require("../../../types/factories");
const helpers_1 = require("./helpers");
const parseFormData = (data) => {
    return [...data].reduce((acc, [key, value]) => {
        if (typeof value === 'string') {
            acc[key] = value;
        }
        return acc;
    }, {});
};
const parseLabels = (labels) => {
    if (!labels)
        return [];
    return [...labels].map(parseToLabel).filter((el) => Boolean(el));
};
const parseToLabel = (label) => {
    const textContent = label.textContent ? (0, helpers_1.cleanText)(label.textContent) : '';
    return {
        id: label.id,
        attributes: parseNodeMap(label.attributes),
        textContent,
    };
};
const parseNodeMap = (nodeMap) => {
    return Array.from(nodeMap).reduce((acc, attr) => {
        acc[attr.name] = attr.value;
        return acc;
    }, {});
};
/**
 * Get the element referenced from an type
 */
const getReferencedElement = (el, attr) => {
    const value = el.getAttribute(attr);
    if (!value)
        return undefined;
    return document.getElementById(value) ?? undefined;
};
const parseElement = (el) => {
    const labels = parseLabels(el.labels);
    const labeledBy = getReferencedElement(el, 'aria-labelledby');
    const describedBy = getReferencedElement(el, 'aria-describedby');
    if (labeledBy) {
        const label = parseToLabel(labeledBy);
        labels.unshift(label);
    }
    const base = {
        // adding a bunch of fields that are not on _all_ elements, but are on enough that it's useful to have them here.
        attributes: parseNodeMap(el.attributes),
        classList: [...el.classList],
        id: el.id,
        labels,
        label: labels[0],
        name: el.name,
        nodeName: el.nodeName,
        tagName: el.tagName,
        title: el.title,
        type: el.type,
        value: el.value,
        textContent: (el.textContent && (0, helpers_1.cleanText)(el.textContent)) ?? undefined,
        innerText: (el.innerText && (0, helpers_1.cleanText)(el.innerText)) ?? undefined,
        describedBy: (describedBy && parseToLabel(describedBy)) ?? undefined,
    };
    if (el instanceof HTMLSelectElement) {
        return {
            ...base,
            selectedOptions: [...el.selectedOptions].map((option) => ({
                value: option.value,
                label: option.label,
            })),
            selectedIndex: el.selectedIndex,
        };
    }
    else if (el instanceof HTMLInputElement) {
        return {
            ...base,
            checked: el.checked,
        };
    }
    else if (el instanceof HTMLMediaElement) {
        return {
            ...base,
            currentSrc: el.currentSrc,
            currentTime: el.currentTime,
            duration: el.duration,
            ended: el.ended,
            muted: el.muted,
            paused: el.paused,
            playbackRate: el.playbackRate,
            readyState: el.readyState,
            src: el.src,
            volume: el.volume,
        };
    }
    else if (el instanceof HTMLFormElement) {
        return {
            ...base,
            innerText: undefined,
            textContent: undefined,
            formData: parseFormData(new FormData(el)),
        };
    }
    return base;
};
exports.parseElement = parseElement;
class ClickSignalsGenerator {
    constructor() {
        this.id = 'click';
    }
    register(emitter) {
        const handleClick = (ev) => {
            const target = ev.target;
            if (!target)
                return;
            const el = this.getClosestClickableElement(target);
            if (el) {
                emitter.emit((0, factories_1.createInteractionSignal)({
                    eventType: 'click',
                    target: (0, exports.parseElement)(el),
                }));
            }
        };
        document.addEventListener('click', handleClick, true);
        return () => document.removeEventListener('click', handleClick);
    }
    getClosestClickableElement(el) {
        // if you click on a nested element, we want to get the closest clickable ancestor. Useful for things like buttons with nested value or images
        const selector = [
            'button',
            'a',
            'option',
            '[role="button"]',
            '[role="link"]',
            '[role="menuitem"]',
            '[role="menuitemcheckbox"]',
            '[role="menuitemradio"]',
            '[role="tab"]',
            '[role="option"]',
            '[role="switch"]',
            '[role="treeitem"]',
        ].join(', ');
        return el.closest(selector);
    }
}
exports.ClickSignalsGenerator = ClickSignalsGenerator;
class FormSubmitGenerator {
    constructor() {
        this.id = 'form-submit';
    }
    register(emitter) {
        const handleSubmit = (ev) => {
            const target = ev.target;
            if (!target)
                return;
            // reference to the form element that the submit event is being fired at
            const submitter = ev.submitter;
            // If the form is submitted via JavaScript using form.submit(), the submitter property will be null because no specific button/input triggered the submission.
            emitter.emit((0, factories_1.createInteractionSignal)({
                eventType: 'submit',
                target: (0, exports.parseElement)(target),
                submitter: submitter ? (0, exports.parseElement)(submitter) : undefined,
            }));
        };
        document.addEventListener('submit', handleSubmit, true);
        return () => document.removeEventListener('submit', handleSubmit);
    }
}
exports.FormSubmitGenerator = FormSubmitGenerator;
const shouldIgnoreElement = (el) => {
    if (el instanceof HTMLInputElement) {
        return el.type === 'password';
    }
    return false;
};
exports.shouldIgnoreElement = shouldIgnoreElement;
class OnNavigationEventGenerator {
    constructor() {
        this.id = 'navigation';
    }
    register(emitter) {
        // emit navigation signal on page load
        emitter.emit((0, factories_1.createNavigationSignal)({
            action: 'pageLoad',
            ...this.createCommonFields(),
        }));
        // emit a navigation signal whenever the URL has changed
        const urlChange = new detect_url_change_1.URLChangeObservable();
        urlChange.subscribe((prevUrl) => emitter.emit((0, factories_1.createNavigationSignal)({
            action: 'urlChange',
            prevUrl,
            ...this.createCommonFields(),
        })));
        return () => {
            urlChange.unsubscribe();
        };
    }
    createCommonFields() {
        return {
            // these fields are named after those from the page call, rather than a DOM api.
            url: location.href,
            path: location.pathname,
            hash: location.hash,
            search: location.search,
            title: document.title,
        };
    }
}
exports.OnNavigationEventGenerator = OnNavigationEventGenerator;
//# sourceMappingURL=dom-gen.js.map