import { SignalsIngestClient } from '../client';
import { getSignalBuffer } from '../buffer';
import { SignalEmitter } from '../emitter';
import { domGenerators } from '../signal-generators/dom-gen';
import { NetworkGenerator } from '../signal-generators/network-gen';
import { registerGenerator } from '../signal-generators/register';
import { AnalyticsService } from '../analytics-service';
import { SignalEventProcessor } from '../processor/processor';
import { Sandbox, SandboxSettings } from '../processor/sandbox';
import { SignalGlobalSettings } from './settings';
import { logger } from '../../lib/logger';
export class Signals {
    constructor(settingsConfig = {}) {
        this.preStartBuffer = [];
        this.cleanup = [];
        this.addToPreStartBuffer = (signal) => {
            this.preStartBuffer.push(signal);
        };
        /**
         * Flush/process any signals that were emitted before the start method was called.
         */
        this.flushPreStartBuffer = (processor) => {
            logger.debug(`Flushing ${this.preStartBuffer.length} events in pre-start buffer`, this.preStartBuffer);
            this.signalEmitter.unsubscribe(this.addToPreStartBuffer);
            this.preStartBuffer.forEach(async (signal) => {
                void processor.process(signal, await this.buffer.getAll());
            });
            this.preStartBuffer = [];
        };
        this.globalSettings = new SignalGlobalSettings(settingsConfig);
        this.signalEmitter = new SignalEmitter();
        this.signalsClient = new SignalsIngestClient(this.globalSettings.ingestClient);
        this.buffer = getSignalBuffer(this.globalSettings.signalBuffer);
        this.signalEmitter.subscribe((signal) => {
            void this.signalsClient.send(signal);
            void this.buffer.add(signal);
        });
        this.signalEmitter.subscribe(this.addToPreStartBuffer);
        void this.registerGenerator([
            ...domGenerators,
            new NetworkGenerator(this.globalSettings.network),
        ]);
    }
    /**
     * Does the following:
     * - Sends any queued signals to the server.
     * - Augments the analytics client to transform events -> signals
     * - Registers custom signal generators.
     */
    async start(analytics) {
        const analyticsService = new AnalyticsService(analytics);
        analyticsService.instance.on('reset', () => {
            this.clearStorage();
        });
        this.globalSettings.update({
            edgeFnDownloadURL: analyticsService.edgeFnSettings?.downloadURL,
            disallowListURLs: [
                analyticsService.instance.settings.apiHost,
                analyticsService.instance.settings.cdnURL,
            ],
            sampleRate: analyticsService.instance.settings.cdnSettings
                .autoInstrumentationSettings?.sampleRate ?? 0,
        });
        const sandbox = new Sandbox(new SandboxSettings(this.globalSettings.sandbox));
        const processor = new SignalEventProcessor(analyticsService.instance, sandbox);
        void this.flushPreStartBuffer(processor);
        this.signalEmitter.subscribe(async (signal) => {
            void processor.process(signal, await this.buffer.getAll());
        });
        await this.registerGenerator([
            analyticsService.createSegmentInstrumentationEventGenerator(),
        ]);
        await this.signalsClient.init({
            writeKey: analyticsService.instance.settings.writeKey,
        });
    }
    stop() {
        this.cleanup.forEach((fn) => fn());
    }
    clearStorage() {
        void this.buffer.clear();
    }
    /**
     * Disable redaction, ingestion of signals, and other logging.
     */
    debug(boolean = true, logLevel) {
        this.globalSettings.signalsDebug.setAllDebugging(boolean);
        logger.enableLogging(logLevel ?? 'info');
    }
    /**
     * Register custom signal generators to emit signals.
     */
    async registerGenerator(generators) {
        this.cleanup.push(await registerGenerator(this.signalEmitter, generators, this.globalSettings));
    }
}
//# sourceMappingURL=signals.js.map